% vms_linker mode --  Special mode to facilitate editing of Linker options files on VMS systems.
% hb, 20-Mar-2015
% 

static variable vms_linker_indent = 4;
static variable vms_linker_continue = 8;

CASE_SEARCH = 0;

static define vms_linker_bol () {
    if (bolp()) {
        newline();
        return 1;
    }
    return 0;
}

static define vms_linker_comment() {
    push_spot();
    bol();
    if (looking_at("!")) {
        pop_spot();
        newline();
        insert ("! ");
        return 1;
    }
    pop_spot();
    return 0;
}

% vms_linker_last_indent should be recursive, but I don't know how to do this in .sl
% so the caller has to call again with greater up_lines
static define vms_linker_last_indent(up_lines) {
    variable lcol;
    push_spot();
    if (up(up_lines)!=up_lines or bolp()) {
        pop_spot();
        return 1;
    }
    eol();
    left(1);
    if (looking_at("-")) {
        right(1);
        % lcol = vms_linker_last_indent ();
        % return 0 so the caller can call this again
        lcol = 0;
        pop_spot();
        return lcol;
    }
    down_1();
    if (looking_at("!")) {
        eol();
        % lcol = vms_linker_last_indent ();
        % return 0 so the caller can call this again
        lcol = 0;
        pop_spot();
        return lcol;
    }
    skip_white();
    lcol = what_column();
    pop_spot();
    return lcol;
}

static define vms_linker_this_indent() {
    variable lcol;
    if (bolp())
        return 1;
    push_spot();
    bol();
    if (looking_at("!")) {
        pop_spot();
        return 1;
    }
    skip_white();
    lcol = what_column();
    pop_spot();
    return lcol;
}

static define vms_linker_cont() {
    left(1);
    if (looking_at("-")) {
        variable col, ins;
        right(1);
        col = vms_linker_this_indent();
        if (col<=3)
            col = 1;
        ins = col-1+vms_linker_continue;
        push_spot();
        if (up_1() and not bolp() and 1==left(1) and looking_at("-"))
            ins = col-1;
        pop_spot();
        newline();
        insert_spaces(ins);
        return 1;
    }
    right(1);
    return 0;
}

static define vms_linker_normal () {
    variable col, i;
    col = vms_linker_this_indent();
    push_spot();
    i = 1;
    if (up_1() and not bolp() and 1==left(1) and looking_at("-")) {
        down_1();
        col = vms_linker_last_indent(i);
        while (col==0) {
                ++i;
            col=vms_linker_last_indent(i);
        }
    }
    pop_spot();
    newline();
    insert_spaces(col-1);
}

define vms_linker_newline () {
    trim();
    % at the begin of line just insert a new line
        if (vms_linker_bol())
        return;
    % continue a full line comment (!...)
    if (vms_linker_comment())
        return;
    if (vms_linker_cont())
        return;
    vms_linker_normal();
}

create_syntax_table("vms_linker");

#ifdef HAS_DFA_SYNTAX
%%% DFA_CACHE_BEGIN %%%
private define setup_dfa_callback (name) {
%    dfa_enable_highlight_cache("vms_linker.dfa", name);
    dfa_define_highlight_rule("!.*$", "comment", name);
    dfa_define_highlight_rule("\"[^\"]*\"", "string", name);
    dfa_define_highlight_rule("/[a-zA-Z][_a-zA-Z0-9\\-]*", "preprocess", name);
    dfa_define_highlight_rule("^[\t ]*[pP][sS][eE][cC][tT]", "keyword0", name);
    dfa_define_highlight_rule("[a-zA-Z\\$][a-zA-Z\\$0-9_;:\\[\\]\\<\\>\\.\\$\\-]*", "Knormal", name);
    dfa_define_highlight_rule("[0-9]+", "number", name);
    dfa_define_highlight_rule("%[oO][0-7]+", "number", name);
    dfa_define_highlight_rule("%[dD][0-9]+", "number", name);
    dfa_define_highlight_rule("%[xX][0-9a-fA-F]+", "number", name);
    dfa_build_highlight_table(name);
}
dfa_set_init_callback (&setup_dfa_callback, "vms_linker");
%%% DFA_CACHE_END %%%
#endif

() = define_keywords_n ("vms_linker", "isdrms", 3, 0);
() = define_keywords_n ("vms_linker", "basecasedzroname", 4, 0);
() = define_keywords_n ("vms_linker", "stack", 5, 0);
() = define_keywords_n ("vms_linker", "symbol", 6, 0);
() = define_keywords_n ("vms_linker", "clustercollectgsmatchisd_maxprotect", 7, 0);
() = define_keywords_n ("vms_linker", "dzro_min", 8, 0);
() = define_keywords_n ("vms_linker", "iosegmentuniversal", 9, 0);
() = define_keywords_n ("vms_linker", "rms_related", 11, 0);
() = define_keywords_n ("vms_linker", "symbol_table", 12, 0);
() = define_keywords_n ("vms_linker", "symbol_vector", 13, 0);
() = define_keywords_n ("vms_linker", "case_sensitiveidentification", 14, 0);
() = define_keywords_n ("vms_linker", "psect_attribute", 15, 0);
() = define_keywords_n ("vms_linker", "rms_related_context", 19, 0);

() = define_keywords_n ("vms_linker", "nord", 2, 1);
() = define_keywords_n ("vms_linker", "absconexegbllclmodovrpicrelshrusrvecwrtyes", 3, 1);
() = define_keywords_n ("vms_linker", "bytedatahexalongnordoctapagequadword", 4, 1);
() = define_keywords_n ("vms_linker", "allocequalnoexenopicnoshrnovecnowrtpsectshortspare", 5, 1);
() = define_keywords_n ("vms_linker", "alwayslequalp0bufs", 6, 1);
() = define_keywords_n ("vms_linker", "globalsnoalloc", 7, 1);
() = define_keywords_n ("vms_linker", "nop0bufsresidentsolitary", 8, 1);
() = define_keywords_n ("vms_linker", "procedure", 9, 1);
() = define_keywords_n ("vms_linker", "universals", 10, 1);
() = define_keywords_n ("vms_linker", "alloc_64bit", 11, 1);
() = define_keywords_n ("vms_linker", "private_data", 12, 1);
() = define_keywords_n ("vms_linker", "noalloc_64bit", 13, 1);
() = define_keywords_n ("vms_linker", "initialization", 14, 1);
() = define_keywords_n ("vms_linker", "private_procedure", 17, 1);
() = define_keywords_n ("vms_linker", "initialization_code", 19, 1);

define_syntax ("([<", ")]>", '(', "vms_linker");

!if (keymap_p ("VMS_LINKER")) {
    make_keymap ("VMS_LINKER");
    definekey ("vms_linker_newline", "^M", "VMS_LINKER");
    definekey ("self_insert_cmd", "^I", "VMS_LINKER");
    definekey ("self_insert_cmd", "\"", "VMS_LINKER");
    definekey ("self_insert_cmd", "'", "VMS_LINKER");
}

define vms_linker_mode () {
    set_syntax_flags ("vms_linker",0x81);
    use_syntax_table("vms_linker");
    use_dfa_syntax(1);
   
    use_keymap ("VMS_LINKER");
    set_mode ("VMS_LINKER", 4);
    run_mode_hooks("vms_linker_mode_hook");
}
